import functools
import inspect
from copy import copy
from typing import Dict, Type, Union, Any, TypeVar, Callable, Set, List

from .interfaces import SpecialDepDefinition
from .exceptions import UnresolvableType, DuplicateDefinition
from .definitions import normalise, Singleton, Construction
from .util.reflection import RETURN_ANNOTATION

UNRESOLVABLE_TYPES = [str, int, float, bool]

X = TypeVar("X")

DepDefinition = Any


class Container:
    _registered_types: Dict[Type, SpecialDepDefinition]
    _explicitly_registered_types: Set[Type]

    def __init__(self):
        self._registered_types = {}
        self._explicitly_registered_types = set()

    def define(self, dep: Union[Type[X], Type], resolver: DepDefinition) -> None:
        if dep in self._explicitly_registered_types:
            raise DuplicateDefinition()
        self._registered_types[dep] = normalise(resolver, self)
        self._explicitly_registered_types.add(dep)

    def resolve(self, dep_type: Type[X], suppress_error=False) -> X:
        try:
            if dep_type in UNRESOLVABLE_TYPES:
                raise UnresolvableType(dep_type)
            registered_type = self._registered_types.get(dep_type, dep_type)
            return self._build(registered_type)
        except UnresolvableType as inner_error:
            if not suppress_error:
                raise UnresolvableType(dep_type) from inner_error
            return None  # type: ignore

    def partial(
        self, func: Callable[..., X], keys_to_skip=None, shared: List[Type] = None
    ) -> Callable[..., X]:
        if shared:
            return self._partial_with_shared_singletons(func, shared)
        spec = inspect.getfullargspec(func)

        def _bind_func():
            bindable_deps = self._infer_dependencies(
                spec, suppress_error=True, keys_to_skip=keys_to_skip or []
            )
            return functools.partial(func, **bindable_deps)

        if inspect.iscoroutinefunction(func):

            async def _async_wrapper(*args, **kwargs):
                return await _bind_func()(*args, **kwargs)  # type: ignore

            return _async_wrapper

        # This function exists so that binding can be used in places
        # that rely on `inspect.is_function` to return True.
        # The output from functools.partial will evaluate to False so we
        # need to wrap the call.
        def _compatibility_wrapper(*args, **kwargs):
            return _bind_func()(*args, **kwargs)

        return _compatibility_wrapper

    def clone(self):
        new_container = Container()
        new_container._registered_types = copy(self._registered_types)

        # Even though the new container has the type definitions we want
        # them all to be overridable so the clone can have updated
        # definitions
        new_container._explicitly_registered_types = set()

        return new_container

    def __getitem__(self, dep: Type[X]) -> X:
        return self.resolve(dep)

    def __setitem__(self, dep: Type, resolver: DepDefinition):
        self.define(dep, resolver)

    def _build(self, dep_type: Any) -> Any:
        if isinstance(dep_type, SpecialDepDefinition):
            return dep_type.get_instance(self._build)
        return self._reflection_build(dep_type)

    def _reflection_build(self, dep_type: Type[X]) -> X:
        spec = inspect.getfullargspec(dep_type.__init__)
        sub_deps = self._infer_dependencies(spec)
        try:
            return dep_type(**sub_deps)  # type: ignore
        except TypeError as type_error:
            raise UnresolvableType(dep_type) from type_error

    def _infer_dependencies(
        self, spec: inspect.FullArgSpec, suppress_error=False, keys_to_skip=None
    ):
        keys_to_skip = keys_to_skip or []
        sub_deps = {
            key: self.resolve(sub_dep_type, suppress_error=suppress_error)
            for (key, sub_dep_type) in spec.annotations.items()
            if key != RETURN_ANNOTATION
            and sub_dep_type != Any
            and key not in keys_to_skip
        }
        filtered_deps = {key: dep for (key, dep) in sub_deps.items() if dep is not None}
        return filtered_deps

    def _partial_with_shared_singletons(
        self, func: Callable[..., X], shared: List[Type]
    ):
        def _cloned_container():
            temp_container = self.clone()
            # For each of the shared dependencies resolve before invocation
            # and replace with a singleton
            for type_def in shared:
                temp_container[type_def] = Singleton(
                    Construction(lambda: self.resolve(type_def))
                )
            return temp_container

        if inspect.iscoroutinefunction(func):

            async def _async_function(*args, **kwargs):
                temp_container = _cloned_container()
                temp_bound_func = temp_container.partial(func, shared=[])
                return await temp_bound_func(*args, **kwargs)

            return _async_function
        else:

            def _function(*args, **kwargs):
                temp_container = _cloned_container()
                temp_bound_func = temp_container.partial(func, shared=[])
                return temp_bound_func(*args, **kwargs)

            return _function



import functools
import inspect
from typing import Union, Type, Optional, Callable, TypeVar, Any

from .exceptions import InvalidDependencyDefinition
from .interfaces import SpecialDepDefinition
from .util.functional import arity

X = TypeVar("X")


class Construction(SpecialDepDefinition[X]):
    constructor: Callable[[], X]

    def __init__(self, constructor):
        self.constructor = constructor

    def get_instance(self, _build_func) -> X:
        call = self.constructor  # type: ignore
        return call()


class Alias(SpecialDepDefinition[X]):
    alias_type: Type[X]

    def __init__(self, alias_type):
        self.alias_type = alias_type

    def get_instance(self, build_func) -> X:
        return build_func(self.alias_type)


class Singleton(SpecialDepDefinition[X]):
    singleton_type: Union[Type[X], Construction[X]]
    _instance: Optional[X]

    def __init__(self, singleton_type: Union[Type[X], Construction[X]]):
        self.singleton_type = singleton_type
        self._instance = None

    def get_instance(self, build_func) -> X:
        if self._has_instance:
            return self._instance  # type: ignore
        return self._set_instance(build_func(self.singleton_type))

    @property
    def _has_instance(self):
        return self._instance is not None

    def _set_instance(self, instance: X):
        self._instance = instance
        return instance


def normalise(resolver: Any, container) -> SpecialDepDefinition:
    if isinstance(resolver, SpecialDepDefinition):
        return resolver
    elif inspect.isfunction(resolver):
        return _build_lambda_constructor(resolver, container)
    elif not inspect.isclass(resolver):
        return Singleton(lambda: resolver)  # type: ignore
    else:
        return Alias(resolver)


def _build_lambda_constructor(resolver: Callable, container) -> Construction:
    artiy = arity(resolver)
    if artiy == 0:
        return Construction(resolver)
    if artiy == 1:
        return Construction(functools.partial(resolver, container))
    raise InvalidDependencyDefinition(f"Arity {arity} functions are not supported")



import pytest

from lagom import Singleton, Construction, Container
from lagom.exceptions import InvalidDependencyDefinition


class MyBasicDep:
    basic_value = 5
    pass


class ComplexDep:
    def __init__(self, some_dep):
        self.value = some_dep.basic_value


def test_lambda_arity_zero_works(container: Container):
    container.define(MyBasicDep, lambda: MyBasicDep())
    assert type(container[MyBasicDep]) == MyBasicDep


def test_lambda_arity_one_is_passed_the_container(container: Container):
    container.define(MyBasicDep, lambda: MyBasicDep())
    container.define(ComplexDep, lambda c: ComplexDep(c[MyBasicDep]))
    assert container[ComplexDep].value == 5


def test_lambda_arity_two_results_in_an_error(container: Container):
    with pytest.raises(InvalidDependencyDefinition):
        container.define(MyBasicDep, lambda _x, _y: MyBasicDep)



import pytest

from lagom import Construction, Container
from lagom.exceptions import DuplicateDefinition


class InitialDep:
    pass


class SomeMockForTesting(InitialDep):
    pass


class SomeOtherMockForTesting(InitialDep):
    pass


def test_container_can_be_cloned_and_maintains_separate_deps(container: Container):
    new_container = container.clone()
    new_container.define(InitialDep, Construction(lambda: SomeMockForTesting()))

    assert isinstance(new_container[InitialDep], SomeMockForTesting)
    assert isinstance(container[InitialDep], InitialDep)


def test_a_cloned_container_can_have_deps_overwritten(container: Container):
    container.define(InitialDep, Construction(lambda: SomeMockForTesting()))
    new_container = container.clone()
    new_container.define(InitialDep, Construction(lambda: SomeOtherMockForTesting()))

    assert isinstance(new_container[InitialDep], SomeOtherMockForTesting)



import inspect

import pytest

from lagom import Container, bind_to_container


class Something:
    pass


def test_partial_application_async_functions_pass_iscoroutinefunction(
    container: Container
):
    @bind_to_container(container)
    async def example_async_function(message: str) -> str:
        return message

    assert inspect.iscoroutinefunction(example_async_function)


def test_partial_application_async_functions_with_shared_pass_iscoroutinefunction(
    container: Container
):
    @bind_to_container(container, shared=[Something])
    async def example_async_function(message: str) -> str:
        return message

    assert inspect.iscoroutinefunction(example_async_function)


@pytest.mark.asyncio
async def test_calling_async_partials_works_as_expected(container: Container):
    @bind_to_container(container)
    async def example_async_function(message: str) -> str:
        return message

    assert await example_async_function("test") == "test"


@pytest.mark.asyncio
async def test_calling_async_partials_works_as_expected_with_shared_too(
    container: Container
):
    @bind_to_container(container, shared=[Something])
    async def example_async_function(message: str) -> str:
        return message

    assert await example_async_function("test") == "test"



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

X = TypeVar("X")


class SpecialDepDefinition(ABC, Generic[X]):
    @abstractmethod
    def get_instance(self, build_func) -> X:
        pass



from typing import Type


class InvalidDependencyDefinition(ValueError):
    pass


class DuplicateDefinition(ValueError):
    pass


class UnresolvableType(ValueError):
    dep_type: str

    def __init__(self, dep_type: Type):
        self.dep_type = dep_type.__name__
        super().__init__(
            f"Unable to construct dependency of type {self.dep_type} "
            "The constructor probably has some unresolvable dependencies"
        )



from abc import ABC
from dataclasses import dataclass

import typing

from lagom import Container, bind_to_container


# Here is an example of some classes your application may be built from


@dataclass
class DiceApiUrl:
    url: str


class RateLimitingConfig:
    pass


class DiceClient(ABC):
    pass


class PlayerDb:
    def __init__(self, some_dsn):
        pass

    def load(self, name: str):
        # Maybe cache what we've loaded. But for how long?
        return "DATA"


class LocalCachingPlayerDb(PlayerDb):
    cache: typing.Dict

    def load(self, name: str):
        # cache what we've loaded FOR EVER
        self.cache[name] = "CACHE"
        return "DATA"


class HttpDiceClient(DiceClient):

    def __init__(self, url: DiceApiUrl, limiting: RateLimitingConfig):
        pass


class Game:
    def __init__(self, dice_roller: DiceClient):
        pass

    def make_move(self, a, b, move):
        pass


# Next we setup some definitions

container = Container()
# We need a specific url
container[DiceApiUrl] = DiceApiUrl("https://roll.diceapi.com")

# Our player DB should be reloaded each time as it caches locally
# we'll share it per request later
container[PlayerDb] = lambda: LocalCachingPlayerDb("blah blah blah")

# Wherever our code wants a DiceClient we get the http one
container[DiceClient] = HttpDiceClient


# If we wanted to build this into a web framework for example
# we can partially bind functions to the container we share the
# PlayerDB so the object can cache data locally but only for the
# lifetime of a single request
@bind_to_container(container, shared=[PlayerDb])
def handle_some_request(request: typing.Dict, game: Game, player_db: PlayerDb):
    player_one = player_db.load(request["player_one"])
    player_two = player_db.load(request["player_two"])
    game.make_move(player_one, player_two, request["move"])


# we can now call the following. the game argument will automagically
# come from the container
handle_some_request(request={"roll_dice": 5})



import pytest

from lagom import Container
from lagom.exceptions import UnresolvableType


class MyMissingDep:
    def __init__(self, _stuff: str):
        pass


class UnfulfilledDeps:
    def __init__(self, _stuff: MyMissingDep):
        pass


@pytest.mark.parametrize("dep", [str, int, float, bool])
def test_simple_objects_cannot_be_resolved(container: Container, dep):
    with pytest.raises(UnresolvableType):
        container.resolve(dep)


def test_raises_error_with_the_dep_that_couldnt_be_built(container):
    with pytest.raises(UnresolvableType):
        container.resolve(MyMissingDep)


def test_raises_error_with_the_dep_that_couldnt_be_built_at_the_top_level(container):
    with pytest.raises(UnresolvableType) as e_info:
        container.resolve(UnfulfilledDeps)
    assert (
        str(e_info.value) == "Unable to construct dependency of type UnfulfilledDeps "
        "The constructor probably has some unresolvable dependencies"
    )



from abc import ABC
from typing import List

import pytest

from lagom import Construction, Container, Alias


class MySimpleAbc(ABC):
    stuff: str = "empty"

    def bloop(self):
        return "beep"


class MySimpleDep(MySimpleAbc):
    stuff: str

    def __init__(self, stuff):
        self.stuff = stuff


class MyMoreComplicatedDep:
    complicated_stuff: str

    def __init__(self, dep: MySimpleAbc):
        self.stuff = dep.stuff


class AnotherAbc(ABC):
    stuff: str = "empty"


class AnotherConcrete(AnotherAbc):
    stuff = "full"


@pytest.fixture
def container_with_abc(container: Container):
    container.define(MySimpleAbc, Construction(lambda: MySimpleDep("hooray")))
    return container


def test_registered_concrete_class_is_loaded(container_with_abc: Container):
    resolved = container_with_abc.resolve(MySimpleAbc)
    assert resolved.stuff == "hooray"


def test_registered_concrete_class_is_used_for_other_objects(
    container_with_abc: Container
):
    resolved = container_with_abc.resolve(MyMoreComplicatedDep)
    assert resolved.stuff == "hooray"


def test_alias_can_be_defined(container_with_abc: Container):
    container_with_abc.define(AnotherAbc, Alias(AnotherConcrete))
    resolved = container_with_abc.resolve(AnotherAbc)
    assert resolved.stuff == "full"
